# EnSimpleStaging  |  Copyright(C), 2004, Enfold Systems, LLC
# see LICENSE.txt for details

from __future__ import nested_scopes

from DateTime import DateTime
from BTrees.IOBTree import IOBTree
from BTrees.Length import Length
from difflib import SequenceMatcher
from sets import Set

from AccessControl import ClassSecurityInfo
from Products.Archetypes.public import *
from Products.Archetypes.utils import findDict
from Products.EnSimpleStaging.config import PROJECTNAME
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.CMFCorePermissions import View, ModifyPortalContent
from Products.CMFStaging.StagingTool import StageObjects
import zLOG

class StagingArea(BaseFolder):
    """ A simple workspace implementation """

    security = ClassSecurityInfo()
    portal_type = 'StagingArea'
    meta_type = archetype_name = 'Staging Area'

    actions = ({'id': 'view',
                'name': 'View',
                'action': 'string:${object_url}/stagingarea_view',
                'permissions': (View,)},
               )

    _stage_info = None
    _si_length = None

    schema = BaseFolder.schema + Schema((
        StringField(
        'stagePath',
        widget=StringWidget(label='Deployment Path',
                            description=(
        'Path relative to the portal object, '
        'where content from the PUBLIC '
        'folder will be published into on '
        'deployment.  i.e. public_website/sales'
        'NOTICE: the destination must exist. '
        'An empty value means this workspace is disabled and '
        'will not deploy at all.'),
                            ),
        ),))

    def _resetAnnotations(self):
        self._stage_info = IOBTree()
        self._si_length = Length()

    def initializeArchetype(self, **kwargs):
        BaseFolder.initializeArchetype(self, **kwargs)
        sandboxes = [('public','Public Node'), ('private','Private Sandbox')]
        for id, title in sandboxes:
            if not hasattr(self, id):
                self.invokeFactory(Sandbox.portal_type, id=id, title=title)
        self._resetAnnotations()

    security.declareProtected(ModifyPortalContent, 'setStagePath')
    def setStagePath(self, value):
        """ Sets the destination stage
        """
        # change the destination stage information
        field = self.Schema()['stagePath']
        value = str(value) # in case of unicode/marshall
        StringField.set(field, self, value)
        if not value:
            # the stage has been disabled for deployment
            return
        wt = getToolByName(self, 'portal_workspaces')
        wt.checkDuplicateStages()

    def _getSourceStageName(self):
        return 'src_%s' % self.UID()

    security.declarePrivate('getSourceStageInfo')
    def getSourceStageInfo(self):
        if not getattr(self, 'public', None):
            # no public sandbox set, it's likely we're being
            # created.
            return None
        ut = getToolByName(self, 'portal_url')
        title = self.title_or_id() + " (source)"
        return (self._getSourceStageName(), # id
                title, # title (duh :-)
                ut.getRelativeContentURL(self.public), # path
                )

    security.declarePrivate('getDestinationStageInfo')
    def getDestinationStageInfo(self):
        field = self.Schema()['stagePath']
        path = StringField.get(field, self)
        if not path:
            # no dest. path initialized in this object
            return None
        portal = getToolByName(self, 'portal_url').getPortalObject()
        title = self.title_or_id() + " (destination)"
        return ('dst_%s' % self.UID(), # id
                title,
                path,
                )

    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        BaseFolder.manage_afterAdd(self, item, container)
        wt = getToolByName(self, 'portal_workspaces')
        wt.checkDuplicateStages()

    def _globalLabel(self, label):
        # global labels must be unique among all labels in a site
        return '%s-%s' % (self.UID(), label)

    security.declareProtected(StageObjects, 'getNextLabel')
    def getNextLabel(self):
        return "%s" % self._si_length()

    def getLastChangedPaths(self):
        # return paths for objects from the catalog
        # that were changed since the last publication or tagging
        annotIndex, annot = self.getLastReversibleAnnotation()
        lastPubDate = annot['date']
        pubNode = self.public
        return pubNode.ess_getLastChangedPathsSince(lastPubDate)


    def getLastMovements(self):
        index, annotLast = self.getLastReversibleAnnotation()
        index, annotPrevious = self.getLastReversibleAnnotation(before=index)
        wt = getToolByName(self, 'portal_workspaces')
        stage = wt.getStageOf(self.public)
        historyIdMapPrevious = wt.getPath2HistoryIdMap(stage,
                                                       annotPrevious['args']['label'])
        historyIdMapLast = wt.getPath2HistoryIdMap(stage,
                                                   annotLast['args']['label'])
        differ = SequenceMatcher(None, historyIdMapPrevious, historyIdMapLast)
        changes = differ.get_opcodes()
        removes, adds = interpretOpcodes(changes,
                                         historyIdMapPrevious, historyIdMapLast)
        removes = ["/".join(path) for path, histId in removes]
        adds = ["/".join(path) for path, histId in adds]
        return removes, adds

    def getLastMovementsAndModifications(self):
        # an object is considered removed if a (path, history_id) tuple
        # is present in the previous label but not on this one.
        # It's considered added if the reverse happens
        # It's considered changed if the (path, history_id) remains
        # between labels, but the version_id changes.
        index, annotLast = self.getLastReversibleAnnotation()
        index, annotPrevious = self.getLastReversibleAnnotation(before=index)
        wt = getToolByName(self, 'portal_workspaces')
        stage = wt.getStageOf(self.public)
        vInfoMapPrevious = wt.getPath2VersionInfoMap(stage,
                                                     annotPrevious['args']['label'])
        historyIdMapPrevious = [(path, history_id)
                                for path, history_id, version_id
                                in vInfoMapPrevious]
        himPreviousSet = Set(historyIdMapPrevious)
        vInfoMapLast = wt.getPath2VersionInfoMap(stage,
                                                 annotLast['args']['label'])
        historyIdMapLast = [(path, history_id)
                            for path, history_id, version_id in vInfoMapLast]
        himLastSet = Set(historyIdMapLast)
        # calculate the movements
        removes = himPreviousSet - himLastSet
        adds = himLastSet - himPreviousSet
        # now calculate the changes:
        # & -> intersection for sets
        # XXX unused variable?
        # remainderPaths = [path for path, history_id
        #                   in himPreviousSet & himLastSet]
        versionIdMapPrevious = [(path, version_id)
                                for path, history_id, version_id
                                in vInfoMapPrevious]
        vimPreviousSet = Set(versionIdMapPrevious)
        versionIdMapLast = [(path, version_id)
                            for path, history_id, version_id
                            in vInfoMapLast]
        vimLastSet = Set(versionIdMapLast)
        modifications = (vimLastSet - vimPreviousSet)
        removes, adds, modifications = [_getPathAndSort(seq) for seq
                                        in removes, adds, modifications]
        # modifications should not have dupes with adds and removes
        # XXX This breaks the testChangeSetsPartialDeployment
        modifications = list(Set(modifications) - Set(adds) - Set(removes))
        return dict(removes=removes, adds=adds, modifications=modifications)

    security.declareProtected(StageObjects, 'publishLastChanged')
    def publishLastChanged(self, label=None, message=None):
        get_transaction().note("incremental deployment of stage '%s'" % "/".join(self.getPhysicalPath()))
        # first, let's get the deployment information
        wt = getToolByName(self, 'portal_workspaces')
        si = wt.getStagesInfo()
        to_stage = si.get(self.getStagePath())
        __traceback_info__=(self.getId(), 'destination stage is ' + str(to_stage), si)
        if to_stage is None:
            raise ValueError, "Destination stage is not set. Cannot proceed."
        to_stages = [to_stage[0]]
        # now let's locate the changed objects...
        changedPaths = self.getLastChangedPaths()
        # and check them in
        wt.checkInObjects([self.unrestrictedTraverse(path)
                           for path in changedPaths],
                          message)
        # now, let's snapshot the current state so that we can
        # revert to it later, and calculate the moved objects
        # (removes, copies, renames)
        if not label:
            label = self.getNextLabel()
        globalLabel = self._globalLabel(label)
        args = {'label':globalLabel,
                'message':message,
                'to_stage':to_stage,
                'to_stage_path':self.getStagePath()}
        info = {'operation': 'partialPublish+tag',
                'args':args,
                'count':len(changedPaths), # only an aproximation
                'label': label}
        self.saveStage(label, message, info)
        # now let's get the movements
        removes, adds = self.getLastMovements()
        # let's remove what's in the way...
        self._removeFromDestination(removes)
        # ...find the duplicates between the changed objects and the
        # moved/copied ones
        pubPath = "/".join(self.public.getPhysicalPath())
        adds = [pubPath + '/' + path
                for path in adds]
        addSet = Set(adds)
        addSet.union_update(changedPaths)
        # and finally, let's updated the newly added and changed
        # objects
        paths = [path.split('/') for path in addSet]
        paths.sort()
        objs = [self.unrestrictedTraverse(path) for path in paths]
        wt.publishObjects(objs, to_stages)

    def _removeFromDestination(self, paths):
        # remove a bunch of relative paths from the destination stage
        root = getToolByName(self, 'portal_url').getPortalObject()
        dest_stage = root.unrestrictedTraverse(self.getStagePath())
        marker = object()
        for path in paths:
            obj = dest_stage.unrestrictedTraverse(path, marker)
            if obj is not marker:
                obj.aq_parent.manage_delObjects([obj.getId()])


    security.declareProtected(StageObjects, 'publishStage')
    def publishStage(self, label=None, message=None, wipe=True, hardWipe=False):
        """Publish this stage to the targeted deployment stage.
        """
        get_transaction().note("full deployment of stage '%s'" % "/".join(self.getPhysicalPath()))
        wt = getToolByName(self, 'portal_workspaces')
        si = wt.getStagesInfo()
        to_stage = si.get(self.getStagePath())
        __traceback_info__=(self.getId(), 'destination stage is ' + str(to_stage), si)
        if to_stage is None:
            raise ValueError, "Destination stage is not set. Cannot proceed."
        if not label:
            label = self.getNextLabel()

        globalLabel = self._globalLabel(label)
        to_stages = [to_stage[0]]
        source = self.public
        from_stage = wt.getStageOf(source)
        assert from_stage is not None
        wt.recursivePublish(source, from_stage, to_stages, label=globalLabel, message=message, wipe=wipe, hardWipe=hardWipe)
        args = {'label':globalLabel,
                'message':message,
                'to_stage':to_stage,
                'to_stage_path':self.getStagePath()}
        info = {'operation': label and 'publish+tag' or 'publish',
                'args':args,
                'count':self.count(),
                'label': label}
        self.annotate(info)

    security.declareProtected(StageObjects, 'saveStage')
    def saveStage(self, label, message=None, info=None):
        """Save and label this stage
        """
        wt = getToolByName(self, 'portal_workspaces')
        __traceback_info__=(self.getId(), )
        globalLabel = self._globalLabel(label)
        source = self.public
        from_stage = wt.getStageOf(source)
        assert from_stage is not None
        wt.recursiveTag(source, from_stage, label=globalLabel,
                        message=message)
        if info is None:
            args = {'label': globalLabel,
                    'message': message}
            info = {'operation': 'tag',
                    'args': args,
                    'count': self.count(),
                    'label': label}
        self.annotate(info)

    security.declarePublic('isReversibleAnnotation')
    def isReversibleAnnotation(self, annotation):
        """check if the annotation represents a reversible tag"""
        # 'tag' must be present in the annotation operation
        # and the global label must be valid
        wt = getToolByName(self, 'portal_workspaces')

        return ('tag' in annotation['operation'] and
                wt.isValidLabel(self._getSourceStageName(),
                                annotation['args']['label']))

    security.declareProtected(StageObjects, 'updateStage')
    def updateStage(self, label):
        """Update stage to a given label.
        """
        # label here is a globalLabel
        if not label:
            raise ValueError, 'Cannot revert to empty label'
        if label not in [annotation['args']['label']
                         for annotation in self.getAnnotations()
                         if self.isReversibleAnnotation(annotation)
                         ]:
            raise ValueError, 'Cannot revert to label "%s"' % label
        wt = getToolByName(self, 'portal_workspaces')
        source = self.public
        wt.recursiveRevertToLabel(source, label)
        args = {'label':label}
        annotation = {'operation': 'update',
                      'count':self.count(),
                      'args':args}
        self.annotate(annotation)

    security.declarePrivate('annotate')
    def annotate(self, info):
        mt = getToolByName(self, 'portal_membership')
        more_info = {'user':mt.getAuthenticatedMember().getId(),
                     'date':DateTime()}
        info.update(more_info)
        self._si_length.change(1)
        next = self._si_length()
        self._stage_info[next] = info

    security.declarePublic('getLastAnnotation')
    def getLastAnnotation(self):
        """ Get the last annotation
        """
        return self._stage_info[self._si_length()]

    security.declarePublic('getLastReversibleAnnotation')
    def getLastReversibleAnnotation(self, before=None):
        """ Get the last annotation that is also a tag operation
        """
        if before is None:
            start = self._si_length()
        else:
            start = before - 1
        for i in xrange(start, 0, -1):
            annot = self._stage_info[i]
            if self.isReversibleAnnotation(annot):
                return i, annot
        raise ValueError, "No remaining reversible operations"

    security.declarePublic('getAnnotations')
    def getAnnotations(self, count=None):
        """ Get a sequence of annotations.

        If the count argument is provided, the 'count'
        last annotations are returned.
        """
        start = 1
        last = self._si_length()
        if count is not None:
            start = last - count
        return [self._stage_info[i] for i in range(start, last+1)]

    security.declarePrivate('count')
    def count(self):
        res = Length()
        def add(obj):
            res.change(1)
        recurseVisit(self.public, add)
        return res()

def recurseVisit(obj, func):
    for item in obj.objectValues():
        func(item)
        if item.aq_explicit.isPrincipiaFolderish:
            recurseVisit(item, func)
    return

def interpretOpcodes(opcodes, ilist, jlist):
    removes = []
    adds = []
    for tag, i1, i2, j1, j2 in opcodes:
        if tag == 'equal':
            # we're not interested in the areas that match
            continue
        if tag == 'delete' or tag == 'replace':
            removes.extend(ilist[i1:i2])
        if tag == 'insert' or tag == 'replace':
            adds.extend(jlist[j1:j2])
    return removes, adds

def _getPathAndSort(sequence):
    sequence = [path for path, dummy in sequence]
    sequence.sort()
    return sequence

registerType(StagingArea, PROJECTNAME)

class Sandbox(OrderedBaseFolder):
    """ Staging Sandbox """

    meta_type = portal_type = archetype_name = 'Sandbox'

registerType(Sandbox, PROJECTNAME)
