##############################################################################
#
# Enfold Stage Deployment - Remote Deployment of Staged Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: subscriber.py 2012 2005-05-31 22:21:48Z sidnei $
"""

from types import ListType, TupleType
from DateTime import DateTime
from Acquisition import aq_inner, aq_parent
from BTrees.OOBTree import OOSet, intersection, difference
from Products.EnSimpleStaging import event
from Products.EnSimpleStaging.content.stage import recurseVisit
from Products.StageDeployment.event import IPublishStageEvent
from Products.StageDeployment.event import IPublishLastChangedEvent
from Products.StageDeployment import formatters
from Products.Archetypes.config import REFERENCE_CATALOG
from Products.Archetypes.utils import shasattr
from Products.Archetypes.debug import log_exc
from Products.CMFCore.utils import getToolByName
from ZConfig.datatypes import SocketAddress

def getParent(item):
    inner = aq_inner(item)
    parent = aq_parent(inner)
    if parent is not None:
        return parent
    # A bug that we haven't been able to track yet
    if hasattr(inner, 'aq_parent'):
        return inner.aq_parent

def getReviewState(item):
    wt = getToolByName(item, 'portal_workflow')
    return wt.getInfoFor(item, 'review_state')

def created(item):
    if not shasattr(item, 'CreationDate'):
        return DateTime()
    dt = DateTime(item.CreationDate())
    return dt

def modified(item):
    if not shasattr(item, 'ModificationDate'):
        return DateTime()
    dt = DateTime(item.ModificationDate())
    return dt

attr_map = {'id': ('getId',),
            'uid': ('UID',),
            'title': ('Title',),
            'description': ('Description',),
            'keywords': ('Subject',),
            'url': ('absolute_url',),
            'credit': ('Contributors',),
            'body':('getText', 'CookedBody', 'getBody',),
            'review_state': (getReviewState,),
            'created': (created,),
            'modified': (modified,),
            'portal_type': ('getPortalTypeName', 'portal_type'),
            'meta_type': ('meta_type',),
            'blob': ('getFile', 'getImageAsFile'),
            }

type_map = {}

# Given a item and a method name, call the
# method to retrieve some value.
# If a tuple of names is provided, return the first name that returns data.
def getValue(item, attrs):
    if type(attrs) not in (ListType, TupleType):
        attrs = (attrs,)
    default = lambda : ''
    value = ''
    for attr_name in attrs:
        if isinstance(attr_name, basestring):
            if not shasattr(item, attr_name):
                continue
            attr = getattr(item, attr_name, default)
        else:
            # It's a function.
            func = attr_name
            def call(item=item):
                return func(item)
            attr = call
        __traceback_info__ = (item, attr_name, attr)
        try:
            if callable(attr):
                value = attr()
            else:
                value = attr
        except:
            log_exc('Failure extracting info %r:%r' % (item, attr_name))
            value = 0
        if value:
            return value
    return value and value or ''

# This is the default value extractor, nothing special about it.
def defaultExtractor(context, obj, obj_attrs, d, fmts, func=getValue):
    for attr in obj_attrs:
        value = func(obj, attr_map.get(attr, attr))
        f = fmts.get(attr, formatters.default)
        d[attr] = f(value)

# Given a set of items and attributes, extract
# the given attributes from each item, using a
# formatting function where it applies.
def extractInfo(context, items, attrs, fmts=None):
    if fmts is None:
        fmts = {}
    info = []
    for item in items:
        d = {}
        extractor = type_map.get(d.get('type'), defaultExtractor)
        extractor(context, item, attrs, d, fmts)
        refs = extractRefs(context, item)
        if refs:
            assert not d.has_key('references'), "Conflict with key 'references'"
            d['references'] = {'subject': item.UID(),
                               'predicates': refs}
        info.append(d)
    return info

def extractRefs(context, item):
    if not shasattr(item, 'UID'):
        return None
    refs = {}
    rt = getToolByName(context, REFERENCE_CATALOG)
    for rname in rt.getRelationships(item):
        sID = item.UID()
        brains = rt._queryFor(sid=sID, relationship=rname)
        if brains:
            rels = {}
            for b in brains:
                rels[b.targetUID] = None
            refs[rname] = rels.keys()
    parent = getParent(item)
    if parent is not None and shasattr(parent, 'UID'):
        refs['contained_into'] = (parent.UID(),)
    return refs

def make_blob(obj):
    if shasattr(obj, 'read'):
        return obj.read()
    # What else?
    return str(obj)

deployment_formatters = {
    'blob':make_blob
    }

def deploymentInfo(context, mode, paths):
    info = []
    items = []
    found = []
    for path in paths:
        # Content is inside the 'public' directory in a staging area.
        content = context.unrestrictedTraverse('public/%s' % path)
        found.append(path)
        items.append(content)
    extracted = extractInfo(context, items,
                            attr_map.keys(), deployment_formatters)
    info.extend(zip(found, (mode,) * len(found), extracted))
    return info

def computeDeployment(stage, adds, updated, removes):
    info = []
    # The operations order shouldn't really matter as we are not
    # making *incremental* changes, but snapshot-based changes, where
    # we have a delta instead of a set of operations performed.  All
    # paths should be ordered from longer to shorter, so that
    # containing folders are created correctly in case they hold any
    # metadata.
    for l in (adds, updated, removes):
        l.sort()
        l.reverse()
    info.extend(zip(removes, ['remove'] * len(removes), [{}] * len(removes)))
    modes = (('add', adds),
             ('update', updated))
    for mode, paths in modes:
        info.extend(deploymentInfo(stage, mode, paths))
    return info

def paths(info):
    return ['/'.join(p) for p in info]

# Event Subscribers
def publishedSubscriber(event):
    from eed.client import Client
    stage = event.obj
    try:
        delta = stage.getLastMovementsAndModifications()
        removes, adds = paths(delta['removes']), paths(delta['adds'])
        updated = paths(delta['modifications'])
    except ValueError:
        # No tag has been found, deploying for the first time it seems.
        base_url = stage.public.absolute_url()
        if not base_url.endswith('/'):
            base_url = '%s/' % base_url
        removes = []
        updated = []
        adds = []
        def collect(obj):
            adds.append(obj.absolute_url().replace(base_url, ''))
        recurseVisit(stage.public, collect)
    info = computeDeployment(stage, adds, updated, removes)
    if not info:
        return
    addr = SocketAddress('127.0.0.1:9001').address
    handler = Client(addr)
    try:
        handler.deploy(info)
    finally:
        handler.close()

event.subscribe(IPublishLastChangedEvent, publishedSubscriber)
event.subscribe(IPublishStageEvent, publishedSubscriber)
