##############################################################################
#
# Enfold Enterprise Deployment - Remote Deployment of Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: __init__.py 732 2005-01-21 19:43:40Z sidnei $
"""
__metaclass__ = type

import tempfile
import logging
from sets import Set
from sqlobject import connectionForURI, SQLObject, SQLObjectNotFound
from eed.logger import TRACE
from eed.utils import copy_file

RESOURCE_NAME = 'SQLObjectManager'

def log(msg, level=TRACE):
    logger = logging.getLogger('eed.common')
    logger.log(level, msg)

def modified(obj):
    obj.syncUpdate()
    expired(obj)

def expired(obj):
    # Expire object values and re-sync from database. The transaction
    # has either been aborted or committed.
    obj_id = obj.id
    obj.expire()
    try:
        # Calling sync() would commit the changes because expire()
        # doesn't clear _SO_createValues. So we do more or less the
        # same as sync() does by calling _init() and setting _expired
        # to False.
        obj._init(obj_id)
        obj._expired = False
    except SQLObjectNotFound:
        # Object may have been deleted or un-created
        pass

class ConnectionDescriptor:

    def __get__(self, inst, cls=None):
        arg = cls is None and inst or cls
        return TransactionFactory(arg)

    def __set__(self, inst, value):
        # Ignore, so we don't get overriden.
        pass

class TransactionWrapper:

    def __init__(self, txn):
        self.txn = txn
        self.objects = Set()

    def __getattribute__(self, attr):
        if attr in ('txn', 'objects', 'commit',
                    'rollback', 'register', 'unregister'):
            return object.__getattribute__(self, attr)
        return getattr(self.txn, attr)

    def __setattr__(self, attr, value):
        if attr in ('txn', 'objects'):
            return object.__setattr__(self, attr, value)
        return setattr(self.txn, attr, value)

    def commit(self):
        for obj in self.objects:
            # XXX Can't know if modified has been fired yet. Maybe we
            # should *not* fire a modified event here?
            modified(obj)
        self.txn.commit()
        self.objects.clear()

    def rollback(self):
        for obj in self.objects:
            expired(obj)
        self.txn.rollback()
        self.objects.clear()

    def register(self, obj):
        if not isinstance(obj, SQLObject):
            raise ValueError, ("Only objects that implement ISQLObject "
                               "can be registered with this transaction "
                               "manager")
        self.objects.add(obj)

    def unregister(self, obj):
        self.objects.discard(obj)

def TransactionFactory(impl):
    from eed.txn import get_resource, add_resource, active
    uri = impl._connection_string
    rsrc_name = '%s for %s' % (RESOURCE_NAME, uri)
    if get_resource(rsrc_name) is not None:
        txn =  get_resource(rsrc_name).txn
    else:
        connection = txn = connectionForURI(uri)

        # Disable Cache
        connection.doCache = False
        connection.cache.kw['cache'] = False

        if active():
            transaction = TransactionWrapper(connection.transaction())
            add_resource(rsrc_name, SQLObjectManager(transaction))
            txn = transaction
        else:
            log('Creating a connection outside active transaction', level=TRACE)
    return txn

def _set_dirty(self, value):
    conn = self._connection
    if value:
        if hasattr(conn, 'register'):
            conn.register(self)
    else:
        if hasattr(conn, 'unregister'):
            conn.unregister(self)
    self._dirty = value

def _get_dirty(self):
    return getattr(self, '_dirty', False)

dirty = property(_get_dirty, _set_dirty)

class SQLObjectManager:

    def __init__(self, transaction):
        self.txn = transaction

    def tpc_begin(self, transaction, subtransaction=False):
        pass

    def commit(self, subtransaction=False):
        self.txn.commit()
        self.txn._makeObsolete()

    def tpc_vote(self, txn):
        pass

    def abort(self, transaction):
        self.txn.rollback()

    tpc_abort = abort

    def tpc_finish(self, transaction):
        pass

    def sortKey(self):
        return 0

    def beforeCompletion(self, txn):
        pass

    def afterCompletion(self, txn):
        from transaction._transaction import Status
        if not txn.status == Status.COMMITTED:
            self.abort(txn)

class TempFile:

    def store(self, stream):
        out, name = tempfile.mkstemp()
        copy_file(stream, out)
        return name

class CommandManager(list):

    def tpc_begin(self, transaction, subtransaction=False):
        for command in self:
            command.prepare()

    def commit(self, subtransaction=False):
        command = None
        for command in self:
            command.execute()

    def tpc_vote(self, txn):
        pass

    def abort(self, transaction):
        command = None
        while self:
            command = self.pop()
            command.revert()
            command.cleanup()

    tpc_abort = abort

    def tpc_finish(self, transaction):
        while self:
            command = self.pop()
            command.cleanup()

    def sortKey(self):
        return 0

    def beforeCompletion(self, txn):
        pass

    def afterCompletion(self, txn):
        from transaction._transaction import Status
        if not txn.status == Status.COMMITTED:
            self.abort(txn)
