##############################################################################
#
# Enfold Enterprise Deployment - Remote Deployment of Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: __init__.py 732 2005-01-21 19:43:40Z sidnei $
"""

import os
import os.path
import tempfile
import eed
import eed.common
import eed.command
import ZConfig
from cStringIO import StringIO

from eed.marshall import xmlrpclibMarshaller
from eed.marshall import BlobMarshaller

_prefix = os.path.dirname(os.path.abspath(__file__))
config_schema_path = os.path.join(_prefix, 'config.xml')
_config_schema = None

config = None

def getConfiguration():
    return getConfig()

def setConfiguration(_config):
    global config
    config = _config

def getConfigSchema():
    global _config_schema
    if _config_schema is None:
        _config_schema = ZConfig.loadSchema(config_schema_path)
    return _config_schema

DEFAULT_CONFIG = """
<config>
  address 9001
</config>
"""

def getConfig(path=None):
    global config
    if config is None:
        schema = getConfigSchema()
        if path is None:
            sio = StringIO(DEFAULT_CONFIG)
            conf, handler = ZConfig.loadConfigFile(schema, sio)
        else:
            conf, handler = ZConfig.loadConfig(schema, path)
        config = conf.config
    return config

def octal_mode(value):
    return int(value, 8)

def Command(section):
    command = eed.command.Command(section.path)
    command.name = '.'.join(section.path.split(os.path.sep))
    return command

def Component(section):
    # .name is resolved to a class through .importable_name
    factory = section.name
    component = factory()
    return component

class _Pipeline(list):

    def __init__(self, name, items):
        self.name = name
        list.__init__(self, items)

    def __repr__(self):
        name = self.name
        return '<Pipeline for %s at %x>' % (name, id(self))

def Pipeline(value):
    comps = value.components
    return _Pipeline(value._name, value.components)

def DataStorage(section=None):
    from eed.storage.fs import DataStorage as Storage
    fs = Storage()
    backup_path = section and section.backup_path or None
    if backup_path is None:
        backup_path = tempfile.mkdtemp()
    fs.backup_path = backup_path
    return fs

def BlobStorage(section=None):
    from eed.storage.fs import BlobStorage as Storage
    fs = Storage()
    backup_path = section and section.backup_path or None
    if backup_path is None:
        backup_path = tempfile.mkdtemp()
    fs.backup_path = backup_path
    return fs

def FilePerms(section=None, storage_name=None):
    from eed.storage.fs import FilePerms as Perms
    perms = Perms()
    perms.file_mode = section and section.file_mode or octal_mode('0644')
    perms.directory_mode = (section and section.directory_mode
                            or octal_mode('0755'))
    perms.storage_name = (section and section.storage_name
                          or storage_name)
    assert perms.storage_name, 'storage name required'
    perms.name = '%s-perms' % perms.storage_name
    return perms

def URLRewrite(section=None, prefix=None, base_url=None, engine=None):
    from eed.storage.url import URLRewrite as Rewrite
    rewrite = Rewrite()
    rewrite.base_url = section and section.base_url or base_url
    engine = section and section.engine or engine
    assert engine is not None, 'engine is required, got: %r' % engine
    rewrite.engine = engine()
    rewrite.prefix = section and section.prefix or prefix
    return rewrite

def SQLObjectMeta(section=None):
    from sqlobject import connectionForURI
    from sqlobject.dbconnection import TheURIOpener
    from eed.metadata._sqlobject import SQLObjectMeta as Metadata
    from eed.metadata._sqlobject import Meta
    connection_string = section and section.connection_string or None
    if connection_string is None:
        connection_string = 'sqlite:///tmp/test.sqlite'
    impl = section and section.implementation or None
    if impl is None:
        impl = Meta
    meta = Metadata()

    # Create tables if not existing
    conn = connectionForURI(connection_string)
    impl._lazyUpdate = False
    impl._connection = conn
    impl.createTable(ifNotExists=True)

    # Now setup a descriptor for the connection so we always get a
    # fresh transaction.
    impl._connection_string = connection_string
    impl._connection = eed.common.ConnectionDescriptor()
    impl.dirty = eed.common.dirty

    meta.impl = impl
    return meta

def ArchetypesTripleStorage(section=None):
    from sqlobject import connectionForURI
    from sqlobject.dbconnection import TheURIOpener
    from rdflib import Graph
    from eed.storage.triplestore import Backend, DumpGraph
    from eed.storage.triplestore import ArchetypesTripleStorage as TS
    from eed.storage.triplestore import AT_UID_NS ,AT_REL_NS
    from eed.storage.triplestore import EED_NS, EED_REL_NS
    connection_string = section and section.connection_string or None
    if connection_string is None:
        connection_string = 'sqlite:///tmp/test.sqlite'
    target_path = section and section.target_path or None
    backup_path = section and section.backup_path or None
    if not ((target_path and backup_path) or
            (not target_path and not backup_path)):
        raise ValueError, ('target_path and backup_path must be both '
                           'informed or both empty: %s, %s' %
                           (target_path, backup_path))
    ts = TS()
    ts.graph = Graph(backend='eed-triple-store-backend')
    if target_path:
        ts.dump = DumpGraph(target_path, ts.graph, backup_path)

    # Create tables if they don't exist yet.
    conn = connectionForURI(connection_string)
    for impl in (ts.triples, ts.resources,
                 ts.literals, ts.ns, ts.prefix):
        impl._lazyUpdate = False
        impl._connection = conn
        impl.createTable(ifNotExists=True)

    # Bind prefixes to namespaces
    ts.graph.bind('atuid', AT_UID_NS[''])
    ts.graph.bind('atrel', AT_REL_NS[''])
    ts.graph.bind('eed', EED_NS[''])
    ts.graph.bind('eedrel', EED_REL_NS[''])

    for impl in (ts.triples, ts.resources,
                 ts.literals, ts.ns, ts.prefix):
        # Now setup a descriptor for the connection so we always get a
        # fresh transaction.
        impl._connection_string = connection_string
        impl._connection = eed.common.ConnectionDescriptor()
        impl.dirty = eed.common.dirty

    return ts

# Some defaults.
DEFAULT_MARSHALL = [xmlrpclibMarshaller(), BlobMarshaller()]
DEFAULT_STORAGE = [DataStorage(), FilePerms(storage_name='data-storage'),
                   BlobStorage(), FilePerms(storage_name='blob-storage')]
DEFAULT_METADATA = []
DEFAULT_BEFORE_HOOKS = [eed.command.before_hooks_queue]
DEFAULT_AFTER_HOOKS = [eed.command.after_hooks_queue]

def root_config(section):
    if section.target_path is None:
        section.target_path = tempfile.mkdtemp()
    if section.marshall is None:
        section.marshall = _Pipeline('marshall', DEFAULT_MARSHALL)
    if section.storage is None:
        storage = DEFAULT_STORAGE + [ArchetypesTripleStorage()]
        section.storage = _Pipeline('storage', storage)
    if section.metadata is None:
        metadata = DEFAULT_METADATA + [SQLObjectMeta()]
        section.metadata = _Pipeline('metadata', metadata)

    # Always register default before hooks
    if section.before_hooks is None:
        section.before_hooks = _Pipeline('before_hooks', DEFAULT_BEFORE_HOOKS)
    else:
        for hook in DEFAULT_BEFORE_HOOKS:
            if hook not in section.before_hooks:
                section.before_hooks.append(hook)

    # Always register default after hooks
    if section.after_hooks is None:
        section.after_hooks = _Pipeline('after_hooks', DEFAULT_AFTER_HOOKS)
    else:
        for hook in DEFAULT_AFTER_HOOKS:
            if hook not in section.after_hooks:
                section.after_hooks.append(hook)
    return section

# A datatype that converts a Python dotted-path-name to an object

def importable_name(name):
    if name.startswith('.'):
        name = eed.config.__name__ + name
    try:
        components = name.split('.')
        start = components[0]
        g = globals()
        package = __import__(start, g, g)
        modulenames = [start]
        for component in components[1:]:
            modulenames.append(component)
            try:
                package = getattr(package, component)
            except AttributeError:
                n = '.'.join(modulenames)
                package = __import__(n, g, g, component)
        return package
    except ImportError:
        raise ValueError, (
            'The object named by "%s" could not be imported' %  name )

# A datatype that ensures that a dotted path name can be resolved but
# returns the name instead of the object

def python_dotted_path(name):
    ob = importable_name(name) # will fail in course
    return name

def existing_executable(name):
    name = os.path.abspath(os.path.expanduser(name))
    if not os.path.exists(name):
        raise ValueError, 'File does not exists: %s' % name
    if not os.access(name, os.X_OK):
        raise ValueError, 'Not an executable file: %s' % name
    return name
