##############################################################################
#
# Enfold Enterprise Deployment - Remote Deployment of Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: __init__.py 732 2005-01-21 19:43:40Z sidnei $
"""

import os
import sys
import time
import tempfile
import logging
from sqlobject import connectionForURI

from eed import txn
from eed.config import getConfiguration
from eed.utils import make_target_path

logger = logging.getLogger('eed.server')

def make_time():
    return str(long(time.mktime(time.gmtime())))

class Handler:

    def deploy(self, struct):
        config = getConfiguration()
        # Setup some environment variables
        os.environ['DEPLOYMENT_DIRECTORY'] = config.target_path
        os.environ['DEPLOYMENT_START'] = make_time()
        try:
            txn.begin()
        except:
            logger.exception('Exception during begin')
            raise
        _txn = txn.get()
        try:
            self.beforeDeployment()
            for key, op, data in struct:
                meth = getattr(self, 'handle_%s' % op, None)
                if meth is not None:
                    meth(key, data)
                else:
                    self.default_handle(op, key, data)
            os.environ['DEPLOYMENT_FINISH'] = make_time()
            self.afterDeployment()
        except:
            logger.exception('Exception ocurred, aborting')
            try:
                txn.abort()
                txn.free(txn.get())
            except:
                logger.exception('Exception during abort')
                raise
            raise
        try:
            txn.commit()
        except:
            # XXX If an exception happens in tpc_prepare the
            # transaction doesn't call free. We shouldn't need to
            # reach into _txn to check this.
            if txn.manager._txn is not None:
                txn.free(_txn)
            logger.exception('Exception during commit')
            raise
        return

    def default_handle(self, op, key, data):
        pass

    def beforeDeployment(self):
        pass

    def afterDeployment(self):
        pass

    def notifyConnected(self, conn):
        pass

    def notifyDisconnected(self):
        pass

class StorageHandler(Handler):

    def normalize_key(self, key):
        config = getConfiguration()
        return make_target_path(config.target_path, key)

    def beforeDeployment(self):
        config = getConfiguration()
        if not config.before_hooks:
            return
        for hook in config.before_hooks:
            try:
                hook.execute()
            finally:
                if hasattr(hook, 'cleanup'):
                    hook.cleanup()

    def afterDeployment(self):
        config = getConfiguration()
        if not config.after_hooks:
            return
        for hook in config.after_hooks:
            try:
                hook.execute()
            finally:
                if hasattr(hook, 'cleanup'):
                    hook.cleanup()

    def default_handle(self, op, key, data):
        config = getConfiguration()
        original_key = key
        key = self.normalize_key(key)

        # More environment madness
        os.environ['DEPLOYMENT_KEY'] = key
        os.environ['DEPLOYMENT_OPERATION'] = op

        info = {}
        info['marshall'] = minfo = {}
        info['storage'] = sinfo = {}
        info['metadata'] = mdinfo = {}
        info['data'] = data
        info['key'] = key
        info['original_key'] = original_key

        for marshaller in config.marshall:
            if op in ('remove',):
                # Remove is ignored for marshall
                continue
            # 1st step, marshall
            assert not minfo.has_key(marshaller.name), (marshaller.name, minfo)
            minfo[marshaller.name] = marshaller.marshall(key, info)

        for storage in config.storage:
            # 2nd step, store data
            assert not sinfo.has_key(storage.name), (storage.name, sinfo)
            sinfo[storage.name] = getattr(storage, op)(key, info)

        for metadata in config.metadata:
            # 3rd step, metadata
            assert not mdinfo.has_key(metadata.name), (metadata.name, mdinfo)
            mdinfo[metadata.name] = getattr(metadata, op)(key, info)
        return 'OK'
