##############################################################################
#
# Enfold Enterprise Deployment - Remote Deployment of Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: __init__.py 732 2005-01-21 19:43:40Z sidnei $
"""

import os
import sys

if __name__ == '__main__':
    from os.path import dirname
    # XXX sys.path hack
    src_dir = dirname(dirname(__file__))
    if not src_dir in sys.path:
        sys.path.insert(0, src_dir)

import signal
import socket

from eed.handler import StorageHandler
from eed.logger import log, WARNING
from ZEO.zrpc.connection import ManagedServerConnection
from ZEO.zrpc.server import Dispatcher

class Server:
    """The server side implementation of EED.

    The Server is the 'manager' for incoming connections.  Each
    connection is associated with its own Handler instance (defined
    below).
    """

    DispatcherClass = Dispatcher
    HandlerClass = StorageHandler
    ManagedServerConnectionClass = ManagedServerConnection

    def __init__(self, addr):
        """Server constructor.

        Arguments:

        addr -- the address at which the server should listen.  This
            can be a tuple (host, port) to signify a TCP/IP connection
            or a pathname string to signify a Unix domain socket
            connection.  A hostname may be a DNS name or a dotted IP
            address.
        """
        self.addr = addr

        self.dispatcher = self.DispatcherClass(addr, factory=self.new_connection)


    def new_connection(self, sock, addr):
        """Internal: factory to create a new connection.

        This is called by the Dispatcher class in ZEO.zrpc.server
        whenever accept() returns a socket for a new incoming
        connection.
        """
        handler = self.HandlerClass()

        c = self.ManagedServerConnectionClass(sock, addr, handler, self)
        return c

    def close_conn(self, conn):
        pass

    def close_server(self):
        """Close the dispatcher so that there are no new connections.

        This is only called from the test suite, AFAICT.
        """
        self.dispatcher.close()
        # Force the asyncore mainloop to exit by hackery, i.e. close
        # every socket in the map.  loop() will return when the map is
        # empty.
        for s in asyncore.socket_map.values():
            try:
                s.close()
            except:
                pass

class DeploymentServer:

    def __init__(self, options):
        self.options = options

    def main(self):
        self.check_socket()
        self.clear_socket()
        try:
            self.setup_signals()
            self.create_server()
            self.loop_forever()
        finally:
            self.clear_socket()

    def setup_signals(self):
        """Set up signal handlers.

        The signal handler for SIGFOO is a method handle_sigfoo().
        If no handler method is defined for a signal, the signal
        action is not changed from its initial value.  The handler
        method is called without additional arguments.
        """
        if os.name != "posix":
            return
        if hasattr(signal, 'SIGXFSZ'):
            signal.signal(signal.SIGXFSZ, signal.SIG_IGN) # Special case
        init_signames()
        for sig, name in signames.items():
            method = getattr(self, "handle_" + name.lower(), None)
            if method is not None:
                def wrapper(sig_dummy, frame_dummy, method=method):
                    method()
                signal.signal(sig, wrapper)

    def handle_sigterm(self):
        log("terminated by SIGTERM")
        sys.exit(0)

    def handle_sigint(self):
        log("terminated by SIGINT")
        sys.exit(0)

    def handle_sighup(self):
        log("restarted by SIGHUP")
        sys.exit(1)

    def handle_sigusr2(self):
        # XXX this used to reinitialize zLOG. How do I achieve
        #     the same effect with Python's logging package?
        #     Should we restart as with SIGHUP?
        log("received SIGUSR2, but it was not handled!", level=WARNING)

    def check_socket(self):
        if self.can_connect(self.options.family, self.options.address):
            raise ValueError, ("address %s already in use" %
                               repr(self.options.address))

    def can_connect(self, family, address):
        s = socket.socket(family, socket.SOCK_STREAM)
        try:
            s.connect(address)
        except socket.error:
            return 0
        else:
            s.close()
            return 1

    def clear_socket(self):
        if isinstance(self.options.address, type("")):
            try:
                os.unlink(self.options.address)
            except os.error:
                pass

    def create_server(self):
        self.server = Server(
            self.options.address)

    def loop_forever(self):
        import ThreadedAsync
        ThreadedAsync.LoopCallback.loop()

# Signal names

signames = None

def signame(sig):
    """Return a symbolic name for a signal.

    Return "signal NNN" if there is no corresponding SIG name in the
    signal module.
    """

    if signames is None:
        init_signames()
    return signames.get(sig) or "signal %d" % sig

def init_signames():
    global signames
    signames = {}
    for name, sig in signal.__dict__.items():
        k_startswith = getattr(name, "startswith", None)
        if k_startswith is None:
            continue
        if k_startswith("SIG") and not k_startswith("SIG_"):
            signames[sig] = name

if __name__ == '__main__':
    # XXX make DateTime safe for marshall
    # DateTime is stitched from Zope trunk. It's put inside ZEO.zrpc
    # so that it doesn't confuse sqlobject into thinking DateTime is
    # mx.DateTime.
    from ZEO.zrpc.DateTime.DateTime import DateTime
    DateTime.__no_side_effects__ = 1

    import sys
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option('-c', '--config',
                      dest='config_file',
                      help='Configuration to use')
    (options, args) = parser.parse_args()
    if not options.config_file:
        parser.error('Config filename is required. '
                     'Use %s -h to see the available '
                     'options.' % sys.argv[0])
    from eed.config import getConfig
    conf = getConfig(options.config_file)
    DeploymentServer(conf.address).main()
