##############################################################################
#
# Enfold Enterprise Deployment - Remote Deployment of Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: __init__.py 732 2005-01-21 19:43:40Z sidnei $
"""

__metaclass__ = type

import os
import stat
import shutil
import tempfile
import logging
from eed.common import CommandManager
from eed.utils import make_target_path
from eed.exceptions import OSError
from eed.logger import TRACE

RESOURCE_NAME = 'CommandManager'

def directory_creation_mode():
    from eed.config import getConfiguration
    return getConfiguration().directory_creation_mode

def copytree(src, dst):
    if not os.path.exists(dst):
        os.makedirs(dst, mode=directory_creation_mode())
    for fname in os.listdir(src):
        path = os.path.join(src, fname)
        if not os.path.isdir(path):
            shutil.copy(path, dst)
        else:
            copytree(path, os.path.join(dst, fname))

def clean_directory(path):
    if os.listdir(path):
        for f in os.listdir(path):
            try:
                shutil.rmtree(os.path.join(path, f))
            except:
                pass

def clean_empty_dirs(path, root):
    root = root.rstrip(os.path.sep)
    parts = path.split(os.path.sep)
    while parts:
        directory = os.path.sep.join(parts)
        if directory == root:
            break
        if not os.path.isdir(directory):
            parts.pop()
            continue
        if not os.listdir(directory):
            os.rmdir(directory)
        parts.pop()

class Command:

    def __init__(self, payload, target):
        if payload is not None:
            assert os.path.exists(payload), 'Must exist: %s' % payload
        self.payload = payload
        self.target = target
        self.executed = False

    def prepare(self):
        raise NotImplementedError

    def execute(self):
        raise NotImplementedError

    def revert(self):
        raise NotImplementedError

    def cleanup(self):
        raise NotImplementedError

class FSCommand(Command):

    def __init__(self, payload, target, backup_path):
        super(FSCommand, self).__init__(payload, target)
        self.backup_path = backup_path
        self.cleanup_path = None

    def cleanup(self):
        if self.cleanup_path is None:
            return
        if os.path.exists(self.cleanup_path):
            if os.path.isdir(self.cleanup_path):
                clean_directory(self.cleanup_path)
            else:
                os.remove(self.cleanup_path)
        clean_empty_dirs(self.cleanup_path, self.backup_path)

class Add(FSCommand):

    target_must_exist = False

    def prepare(self):
        if not os.path.exists(self.target):
            if self.target_must_exist:
                raise OSError, '%s should exist but does not' % self.target
            else:
                return
        bkp = make_target_path(self.backup_path, self.target)
        self.cleanup_path = bkp
        bkp_parent = os.path.dirname(bkp)
        if not os.path.exists(bkp_parent):
            os.makedirs(bkp_parent, mode=directory_creation_mode())
        if not os.path.isdir(self.target):
            shutil.copy(self.target, bkp)
        else:
            # shutil.copytree requires target dir *not* to exist.
            copytree(self.target, bkp)

    def execute(self):
        dst_folder = os.path.dirname(self.target)
        if not os.path.exists(dst_folder):
            os.makedirs(dst_folder, mode=directory_creation_mode())
        shutil.move(self.payload, self.target)
        assert os.path.exists(self.target)
        self.executed = True

    def revert(self):
        if self.executed:
            # only remove if already executed, otherwise we may end up
            # removing existing stuff from the target repository.
            os.remove(self.target)
        else:
            # payload has not been consumed, remove it
            os.remove(self.payload)

class Modify(Add):

    def revert(self):
        if self.executed:
            # repository has been modified, and a backup
            # might exist. restore the original file.
            restore = True
            source = make_target_path(self.backup_path, self.target)
            if not self.target_must_exist:
                if not os.path.exists(source):
                    restore = False
            if restore:
                dest_folder = os.path.dirname(self.target)
                if not os.path.exists(dest_folder):
                    os.makedirs(dest_folder, mode=directory_creation_mode())
                shutil.move(source, self.target)
        # payload has been consumed if executed
        if self.payload is not None and not self.executed:
            os.remove(self.payload)

class Remove(Modify):

    def execute(self):
        if os.path.isdir(self.target):
            shutil.rmtree(self.target)
        else:
            os.remove(self.target)
        self.executed = True

class BlobStorage:

    def __init__(self):
        self.name = 'blob-storage'
        self.marshaller = 'blob-marshaller'
        self.backup_path = None

    def log(self, msg):
        logger = logging.getLogger('eed.%s' % self.name)
        logger.log(TRACE, msg)

    def _manager(self):
        from eed.txn import get_resource
        assert self.backup_path, 'Must set backup_path'
        manager = get_resource(RESOURCE_NAME, CommandManager)
        return manager
    manager = property(_manager)

    def make_key(self, key, hint, nomangledir=False):
        return key

    def add(self, key, info):
        src = info['marshall'][self.marshaller]
        if src is None:
            # Nothing has been marshalled.
            return
        key = self.make_key(key, hint=src)
        self.log('Adding %s' % key)
        self.manager.append(Add(src, key, self.backup_path))
        return key

    def remove(self, key, src=None):
        key = self.make_key(key, hint=key)
        if not os.path.exists(key):
            # File/Directory doesn't exist, ignore right away.
            return
        self.log('Removing %s' % key)
        self.manager.append(Remove(None, key, self.backup_path))
        return key

    def update(self, key, info):
        src = info['marshall'][self.marshaller]
        if src is None:
            # Nothing has been marshalled.
            return
        key = self.make_key(key, hint=src)
        self.log('Updating %s' % key)
        self.manager.append(Modify(src, key, self.backup_path))
        return key

class DataStorage(BlobStorage):

    def __init__(self):
        super(DataStorage, self).__init__()
        self.name = 'data-storage'
        self.marshaller = 'data-marshaller'

    def make_key(self, key, hint):
        return '%s.data' % key

class ChangePermission(Command):

    def __init__(self, payload, target, file_mode, directory_mode):
        super(ChangePermission, self).__init__(payload, target)
        self.file_mode = file_mode
        self.directory_mode = directory_mode
        self.previous_mode = None

    def prepare(self):
        # XXX This doesn't sound quite right. We may need to move this
        # to a post-commit hook or something later on. Save previous
        # permission, if file exists by this time.
        if not os.path.exists(self.target):
            return
        st = os.stat(self.target)
        self.previous_mode = stat.S_IMODE(st.st_mode)

    def execute(self):
        st = os.stat(self.target)
        self.previous_mode = mode = stat.S_IMODE(st.st_mode)
        new_mode = self.file_mode
        if os.path.isdir(self.target):
            new_mode = self.directory_mode
        # Check if it's not already the mode we want.
        if mode % new_mode:
            if hasattr(os, 'chmod'):
                os.chmod(self.target, new_mode)
                self.executed = True

    def revert(self):
        if self.executed:
            assert self.previous_mode is not None
            if hasattr(os, 'chmod'):
                os.chmod(self.target, self.previous_mode)

    def cleanup(self):
        # Nothing to do here.
        pass

class FilePerms:

    def __init__(self):
        # Dummy values, real values are stitched in by the ZConfig.
        self.name = 'data-file-perms'
        self.storage_name = 'data-storage'
        self.file_mode = None
        self.directory_mode = None

    def _manager(self):
        from eed.txn import get_resource
        assert self.file_mode, 'Must set file_mode'
        assert self.directory_mode, 'Must set directory_mode'
        manager = get_resource(RESOURCE_NAME, CommandManager)
        return manager
    manager = property(_manager)

    def handle(self, key, info):
        key = info['storage'][self.storage_name]
        if key is None:
            # No file has been generated by the storage, do nothing
            return
        cmd = ChangePermission(None, key, self.file_mode, self.directory_mode)
        self.manager.append(cmd)
        return key

    add = update = handle

    def remove(self, key, src):
        # We just ignore removes, as the file is gone anyway
        return key
