##############################################################################
#
# Enfold Enterprise Deployment - Remote Deployment of Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: __init__.py 732 2005-01-21 19:43:40Z sidnei $
"""

__metaclass__ = type

import os
import os.path
from sqlobject import *
from rdflib import Namespace, Literal
from rdflib.plugin import register
from rdflib.backends import _sqlobject as backend, Backend
from eed.config import getConfiguration
from eed.storage.fs import Modify, directory_creation_mode
from eed.utils import make_target_path

AT_UID_NS = Namespace('http://archetypes.org/references/uid/')
AT_REL_NS = Namespace('http://archetypes.org/references/relation/')
EED_NS = Namespace('http://enfoldsystems.com/eed/fskey/')
EED_REL_NS = Namespace('http://enfoldsystems.com/eed/relation/')

class TripleStorage(backend.SQLObject):

    # Override open and close, we never use them.
    def open(self, uri): pass
    def close(self): pass

register('eed-triple-store-backend', Backend,
         'eed.storage.triplestore', 'TripleStorage')

class DumpGraph(Modify):

    target_must_exist = False

    def __init__(self, target, graph, backup_path):
        self.graph = graph
        super(DumpGraph, self).__init__(None, target, backup_path)

    def execute(self):
        if self.executed:
            # Don't execute twice
            return
        dest = make_target_path(self.backup_path, self.target)
        dest_folder = os.path.dirname(dest)
        if not os.path.exists(dest_folder):
            os.makedirs(dest_folder, mode=directory_creation_mode())
        f = open(dest, 'wb')
        self.graph.save(f)
        f.close()
        self.payload = dest
        super(DumpGraph, self).execute()

    def cleanup(self):
        super(DumpGraph, self).cleanup()
        self.executed = False

class ArchetypesTripleStorage:

    name = 'archetypes-triple-storage'
    triples = backend.Triples
    literals = backend.Literals
    ns = backend.Namespaces
    prefix = backend.PrefixNamespace
    resources = backend.Resources
    pkey = 'path'
    pkey_hash = None
    graph = None
    dump = None

    def register(self):
        from eed.command import after_hooks_queue
        # Wake up the connections if they haven't been yet.
        foo = (self.literals._connection,
               self.triples._connection,
               self.resources._connection,
               self.ns._connection,
               self.prefix._connection)
        if self.dump is None:
            return
        after_hooks_queue.queue(self.dump)

    def add(self, key, info):
        self.register()
        data = info['data']
        refs = data.get('references')
        if refs is not None:
            # Refs will have a structure like:
            # {'subject': subject,
            #  'predicates':{'predicate': (object, object)}
            # }
            subject = refs['subject']
            suri = AT_UID_NS[subject]
            for pred, objects in refs['predicates'].iteritems():
                puri = AT_REL_NS[pred]
                for obj in objects:
                    # It's always a URI for now.
                    ouri = AT_UID_NS[obj]
                    self.graph.add((suri, puri, ouri))
            # Insert a triple to map the key to the subject.
            klit = Literal(key)
            puri = EED_REL_NS[self.pkey]
            self.graph.add((suri, puri, klit))

    # Alias update to add.
    update = add

    def remove(self, key, info):
        self.register()
        klit = Literal(key)
        puri = EED_REL_NS[self.pkey]
        suri = None
        for s, p, o in self.graph.triples((backend.Any, puri, klit)):
            # Should have only one
            suri = s
        if suri is None:
            # No triple has been found.
            return

        # Delete forward refs.
        for triple in self.graph.triples((suri, backend.Any, backend.Any)):
            self.graph.remove(triple)

        # Delete backward refs
        for triple in self.graph.triples((backend.Any, backend.Any, suri)):
            self.graph.remove(triple)

    def query(self, expand=False, **kw):
        # Wake up the connections if they haven't been yet.
        foo = (self.literals._connection,
               self.triples._connection,
               self.resources._connection,
               self.ns._connection,
               self.prefix._connection)
        return self.graph.triples((backend.Any, backend.Any, backend.Any))
