##############################################################################
#
# Enfold Enterprise Deployment - Remote Deployment of Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: url.py 2027 2005-06-01 01:04:46Z sidnei $
"""

__metaclass__ = type

import os
import re
from eed.common import CommandManager
from eed.storage.fs import RESOURCE_NAME
from eed.storage.fs import Command
from eed.exceptions import URLRewritingError

KUPU_REGEX = '((?<=href=)|(?<=src=))["\'][^\'"]*?resolveuid/(?P<uid>[A-Fa-f0-9]*?)["\']'

class UIDToURL:
    # Depends on CMFMeta

    uids = re.compile(KUPU_REGEX)

    def _get_meta(self):
        from eed.config import getConfiguration
        from eed.metadata._sqlobject import SQLObjectMeta, CMFMeta
        config = getConfiguration()
        for m in config.metadata:
            if isinstance(m, SQLObjectMeta):
                impl = m.impl
                if impl is CMFMeta:
                    return impl
        return None
    meta = property(_get_meta)

    def rewrite(self, key, prefix, path, base_url):
        inf = open(path, 'rb')
        data = inf.read()
        inf.close()
        matches = self.uids.finditer(data)
        for match in matches:
            uid = match.groupdict()['uid']
            url = match.group()
            fpath = self.resolve(uid, url, key)
            relpath = fpath.replace(prefix, '')
            # XXX This could be made stronger.
            target = '"%s"' % (base_url + relpath)
            data, n = self.uids.subn(target, data, 1)
        outf = open(path, 'wb')
        outf.write(data)
        outf.close()

    def resolve(self, uid, url, key):
        meta = self.meta
        if meta is None:
            raise URLRewritingError, 'Required CMFMeta component not found'
        res = meta.select(meta.q.uid == uid)
        if not res.count():
            self.exception(uid, url, key)
        # Should have only one.
        return iter(res).next().path

    def exception(self, uid, url, key):
        raise URLRewritingError, ('Could not find information '
                                  'about object with UID "%s" '
                                  'to rewrite the URL %s '
                                  'while deploying %s.' %
                                  (uid, url, key))

class RewriteURL(Command):

    def __init__(self, payload, target, key, prefix, base_url, engine):
        super(RewriteURL, self).__init__(payload, target)
        self.prefix = prefix
        self.base_url = base_url
        self.engine = engine
        self.key = key

    def prepare(self):
        assert os.path.exists(self.target), \
               'Target file %s must exist' % self.target

    def execute(self):
        if self.engine is not None:
            self.engine.rewrite(self.key, self.prefix, self.target, self.base_url)

    def revert(self):
        # No revert for this command, sorry.
        pass

    def cleanup(self):
        # Nothing to do here.
        pass

class URLRewrite:

    def __init__(self):
        self.name = 'data-url-rewrite'
        self.marshaller_name = 'data-marshaller'
        self.base_url = None
        self.prefix = None
        self.engine = UIDToURL()

    def _manager(self):
        from eed.txn import get_resource
        manager = get_resource(RESOURCE_NAME, CommandManager)
        return manager
    manager = property(_manager)

    def handle(self, key, info):
        assert self.prefix, 'prefix is required'
        assert self.base_url, 'base_url is required'
        assert self.engine, 'engine is required'
        target = info['marshall'][self.marshaller_name]
        if target is None:
            # No file has been generated by the storage, do nothing
            return
        original_key = info['original_key']
        cmd = RewriteURL(None, target, original_key, self.prefix,
                         self.base_url, self.engine)
        self.manager.append(cmd)
        return target

    add = update = handle

    def remove(self, key, src):
        # We just ignore removes, as the file is gone anyway
        return key
